from itertools import chain, repeat

import gem
import numpy
import sympy
try:
    import symengine
    symbolics = symengine
except ImportError:
    symbolics = sympy
from FIAT.polynomial_set import mis
from FIAT.reference_element import UFCQuadrilateral
from gem.utils import cached_property

from finat.finiteelementbase import FiniteElementBase
from finat.physically_mapped import Citations, DirectlyDefinedElement
from finat.sympy2gem import sympy2gem


class DirectSerendipity(DirectlyDefinedElement, FiniteElementBase):
    def __init__(self, cell, degree):
        if Citations is not None:
            Citations().register("Arbogast2017")

        # These elements only known currently on quads
        assert isinstance(cell, UFCQuadrilateral)

        self._cell = cell
        self._degree = degree
        self._deriv_cache = {}

    @property
    def cell(self):
        return self._cell

    @property
    def complex(self):
        return self._cell

    @property
    def degree(self):
        return self._degree

    @property
    def formdegree(self):
        return 0

    def entity_dofs(self):
        if self.degree == 1:
            return {0: {i: [i] for i in range(4)},
                    1: {i: [] for i in range(4)},
                    2: {0: []}}
        elif self.degree == 2:
            return {0: {i: [i] for i in range(4)},
                    1: {i: [i+4] for i in range(4)},
                    2: {0: []}}
        else:
            return {0: {i: [i] for i in range(4)},
                    1: {i: list(range(4 + i * (self.degree-1),
                                      4 + (i + 1) * (self.degree-1)))
                        for i in range(4)},
                    2: {0: list(range(4 + 4 * (self.degree - 1),
                                      self.space_dimension()))}}

    def space_dimension(self):
        return 4 if self.degree == 1 else (self.degree+1)*(self.degree+2)//2 + 2

    @property
    def index_shape(self):
        return (self.space_dimension(),)

    @property
    def value_shape(self):
        return ()

    @cached_property
    def _basis(self):
        return ds_sym(self.cell.topology, self.degree, sp=symbolics)

    def _basis_deriv(self, xx, alpha):
        key = (tuple(xx), alpha)
        _, _, phis = self._basis
        try:
            return self._deriv_cache[key]
        except KeyError:
            dphi = tuple(diff(phi, xx, alpha) for phi in phis)
            return self._deriv_cache.setdefault(key, dphi)

    def basis_evaluation(self, order, ps, entity=None, coordinate_mapping=None):
        '''Return code for evaluating the element at known points on the
        reference element.

        :param order: return derivatives up to this order.
        :param ps: the point set.
        :param entity: the cell entity on which to tabulate.
        '''
        # Build everything in sympy
        vs, xx, _ = self._basis

        # and convert -- all this can be used for each derivative!
        phys_verts = coordinate_mapping.physical_vertices()

        phys_points = gem.partial_indexed(
            coordinate_mapping.physical_points(ps, entity=entity),
            ps.indices)

        repl = dict((vs[idx], phys_verts[idx])
                    for idx in numpy.ndindex(vs.shape))

        repl.update(zip(xx, phys_points))

        mapper = gem.node.Memoizer(sympy2gem)
        mapper.bindings = repl

        result = {}

        for i in range(order+1):
            alphas = mis(2, i)
            for alpha in alphas:
                dphis = self._basis_deriv(xx, alpha)
                result[alpha] = gem.ListTensor(list(map(mapper, dphis)))

        return result

    def point_evaluation(self, order, point, entity=None, coordinate_mapping=None):
        raise NotImplementedError("Not done yet, sorry!")

    def mapping(self):
        return "physical"


def xysub(x, y):
    return {x[0]: y[0], x[1]: y[1]}


def ds1_sym(ct, *, vs=None, sp=symbolics):
    """Constructs lowest-order case of Arbogast's directly defined C^0 serendipity
    elements, which are a special case.
    :param ct: The cell topology of the reference quadrilateral.
    :param vs: (Optional) coordinates of cell on which to construct the basis.
               If it is None, this function constructs symbols for the vertices.
    :returns: a 3-tuple containing symbols for the physical cell coordinates and the
              physical cell independent variables (e.g. "x" and "y") and a list
              of the four basis functions.
    """
    if vs is None:
        vs = numpy.asarray(list(zip(sp.symbols('x:4'),
                                    sp.symbols('y:4'))))
    else:
        vs = numpy.asarray(vs)

    xx = numpy.asarray(sp.symbols("x,y"))

    ts = numpy.zeros((4, 2), dtype=object)
    for e in range(4):
        v0id, v1id = ct[1][e][:]
        for j in range(2):
            ts[e, :] = vs[v1id, :] - vs[v0id, :]

    ns = numpy.zeros((4, 2), dtype=object)
    for e in (0, 3):
        ns[e, 0] = -ts[e, 1]
        ns[e, 1] = ts[e, 0]

    for e in (1, 2):
        ns[e, 0] = ts[e, 1]
        ns[e, 1] = -ts[e, 0]

    xstars = numpy.zeros((4, 2), dtype=object)
    for e in range(4):
        v0id, v1id = ct[1][e][:]
        xstars[e, :] = (vs[v0id, :] + vs[v1id])/2

    lams = [(xx-xstars[i, :]) @ ns[i, :] for i in range(4)]

    RV = (lams[0] - lams[1]) / (lams[0] + lams[1])
    RH = (lams[2] - lams[3]) / (lams[2] + lams[3])
    Rs = [RV, RH]

    xis = []
    for e in range(4):
        dct = xysub(xx, xstars[e, :])
        i = 2*((3-e)//2)
        j = i + 1
        xi = lams[i] * lams[j] * (1 + (-1)**(e+1) * Rs[e//2]) / lams[i].subs(dct) / lams[j].subs(dct) / 2
        xis.append(xi)

    d = xysub(xx, vs[0, :])
    r = lams[1] * lams[3] / lams[1].subs(d) / lams[3].subs(d)
    d = xysub(xx, vs[2, :])
    r -= lams[0] * lams[3] / lams[0].subs(d) / lams[3].subs(d)
    d = xysub(xx, vs[3, :])
    r += lams[0] * lams[2] / lams[0].subs(d) / lams[2].subs(d)
    d = xysub(xx, vs[1, :])
    r -= lams[1] * lams[2] / lams[1].subs(d) / lams[2].subs(d)
    R = r - sum([r.subs(xysub(xx, xstars[i, :])) * xis[i] for i in range(4)])

    n03 = numpy.array([[0, -1], [1, 0]]) @ (vs[3, :] - vs[0, :])
    lam03 = (xx - vs[0, :]) @ n03
    n12 = numpy.array([[0, -1], [1, 0]]) @ (vs[2, :] - vs[1, :])
    lam12 = (xx - vs[2, :]) @ n12

    phi0tilde = lam12 - lam12.subs({xx[0]: vs[3, 0], xx[1]: vs[3, 1]}) * (1 + R) / 2
    phi1tilde = lam03 - lam03.subs({xx[0]: vs[2, 0], xx[1]: vs[2, 1]}) * (1 - R) / 2
    phi2tilde = lam03 - lam03.subs({xx[0]: vs[1, 0], xx[1]: vs[1, 1]}) * (1 - R) / 2
    phi3tilde = lam12 - lam12.subs({xx[0]: vs[0, 0], xx[1]: vs[0, 1]}) * (1 + R) / 2

    phis = []
    for i, phitilde in enumerate([phi0tilde, phi1tilde, phi2tilde, phi3tilde]):
        phi = phitilde / phitilde.subs({xx[0]: vs[i, 0], xx[1]: vs[i, 1]})
        phis.append(phi)

    return vs, xx, numpy.asarray(phis)


def newton_dd(nds, fs):
    """Constructs Newton's divided differences for the input arrays,
    which may include symbolic values."""
    n = len(nds)
    mat = numpy.zeros((n, n), dtype=object)
    mat[:, 0] = fs[:]
    for j in range(1, n):
        for i in range(n-j):
            mat[i, j] = (mat[i+1, j-1] - mat[i, j-1]) / (nds[i+j] - nds[i])
    return mat[0, :]


def newton_poly(nds, fs, xsym):
    """Constructs  Lagrange interpolating polynomial passing through
    x values nds and y values fs.  Returns a a symbolic object in terms
    of independent variable xsym."""
    coeffs = newton_dd(nds, fs)
    result = coeffs[-1]
    n = len(coeffs)
    for i in range(n-2, -1, -1):
        result = result * (xsym - nds[i]) + coeffs[i]
    return result


def diff(expr, xx, alpha):
    """Differentiate expr with respect to xx.

    :arg expr: symengine/symengine Expression to differentiate.
    :arg xx: iterable of coordinates to differentiate with respect to.
    :arg alpha: derivative multiindex, one entry for each entry of xx
        indicating how many derivatives in that direction.
    :returns: New symengine/symengine expression."""
    if isinstance(expr, sympy.Expr):
        return expr.diff(*(zip(xx, alpha)))
    else:
        return symengine.diff(expr, *(chain(*(repeat(x, a) for x, a in zip(xx, alpha)))))


def dsr_sym(ct, r, *, vs=None, sp=symbolics):
    """Constructs higher-order (>= 2) case of Arbogast's directly defined C^0 serendipity
    elements, which include all polynomials of degree r plus a couple of rational
    functions.
    :param ct: The cell topology of the reference quadrilateral.
    :param vs: (Optional) coordinates of cell on which to construct the basis.
               If it is None, this function constructs symbols for the vertices.
    :returns: a 3-tuple containing symbols for the physical cell coordinates and the
              physical cell independent variables (e.g. "x" and "y") and a list
              of the four basis functions.
    """
    if vs is None:  # do vertices symbolically
        vs = numpy.asarray(list(zip(sp.symbols('x:4'),
                                    sp.symbols('y:4'))))
    else:
        vs = numpy.asarray(vs)
    xx = numpy.asarray(sp.symbols("x,y"))

    ts = numpy.zeros((4, 2), dtype=object)
    for e in range(4):
        v0id, v1id = ct[1][e][:]
        ts[e, :] = vs[v1id, :] - vs[v0id, :]

    ns = numpy.zeros((4, 2), dtype=object)
    for e in (0, 3):
        ns[e, 0] = -ts[e, 1]
        ns[e, 1] = ts[e, 0]

    for e in (1, 2):
        ns[e, 0] = ts[e, 1]
        ns[e, 1] = -ts[e, 0]

    # midpoints of each edge
    xstars = numpy.zeros((4, 2), dtype=object)
    for e in range(4):
        v0id, v1id = ct[1][e][:]
        xstars[e, :] = (vs[v0id, :] + vs[v1id])/2

    lams = [(xx-xstars[i, :]) @ ns[i, :] for i in range(4)]

    # # internal functions
    bubble = numpy.prod(lams)

    if r < 4:
        internal_bfs = []
        internal_nodes = []
    elif r == 4:  # Just one point
        xbar = sum(vs[i, 0] for i in range(4)) / 4
        ybar = sum(vs[i, 1] for i in range(4)) / 4
        internal_bfs = [bubble / bubble.subs(xysub(xx, (xbar, ybar)))]
        internal_nodes = [(xbar, ybar)]
    else:  # build a triangular lattice inside the quad
        dx0 = (vs[1, :] - vs[0, :]) / (r-2)
        dx1 = (vs[2, :] - vs[0, :]) / (r-2)

        # Vertices of the triangle
        v0 = vs[0, :] + dx0 + dx1
        v1 = vs[0, :] + (r-3) * dx0 + dx1
        v2 = vs[0, :] + dx0 + (r-3) * dx1

        # Pardon the fortran, but these are barycentric coordinates...
        bary = numpy.zeros((3,), dtype="object")
        y12 = v1[1] - v2[1]
        x21 = v2[0] - v1[0]
        x02 = v0[0] - v2[0]
        y02 = v0[1] - v2[1]
        det = y12 * x02 + x21 * y02
        delx = xx[0] - v2[0]
        dely = xx[1] - v2[1]
        bary[0] = (y12 * delx + x21 * dely) / det
        bary[1] = (-y02 * delx + x02 * dely) / det
        bary[2] = 1 - bary[0] - bary[1]

        # And this bit directly constructs the Lagrange polynomials
        # of degree r-4 on the triangular lattice inside the triangle.
        # This trick is restricted to equispaced points, but we're on a much
        # lower degree than r.  This bypasses symbolic inversion/etc otherwise
        # required to build the Lagrange polynomials.
        rm4 = r - 4
        lags = []
        internal_nodes = []
        for i in range(rm4, -1, -1):
            for j in range(rm4-i, -1, -1):
                k = rm4 - i - j
                internal_nodes.append((v0 * i + v1 * j + v2 * k)/rm4)
                ii = (i, j, k)
                lag_cur = sp.Integer(1)
                for q in range(3):
                    for p in range(ii[q]):
                        lag_cur *= (rm4 * bary[q] - p) / (ii[q] - p)
                lags.append(lag_cur.simplify())

        internal_bfs = []
        for lag, nd in zip(lags, internal_nodes):
            foo = lag * bubble
            internal_bfs.append(foo / foo.subs(xysub(xx, nd)))

    RV = (lams[0] - lams[1]) / (lams[0] + lams[1])
    RH = (lams[2] - lams[3]) / (lams[2] + lams[3])

    # R for each edge (1 on edge, zero on opposite
    Rs = [(1 - RV) / 2, (1 + RV) / 2, (1 - RH) / 2, (1 + RH) / 2]

    nodes1d = [sp.Rational(i, r) for i in range(1, r)]

    s = sp.Symbol('s')

    # for each edge:
    # I need its adjacent two edges and its opposite edge
    # and its "tunnel R" RH or RV
    # This is very 2d specific.
    opposite_edges = {e: [eother for eother in ct[1]
                          if set(ct[1][e]).intersection(ct[1][eother]) == set()][0]
                      for e in ct[1]}
    adjacent_edges = {e: tuple(sorted([eother for eother in ct[1]
                                       if eother != e
                                       and set(ct[1][e]).intersection(ct[1][eother])
                                       != set()]))
                      for e in ct[1]}

    ae = adjacent_edges
    tunnel_R_edges = {e: ((lams[ae[e][0]] - lams[ae[e][1]])
                          / (lams[ae[e][0]] + lams[ae[e][1]]))
                      for e in range(4)}
    edge_nodes = []
    for ed in range(4):
        ((v0x, v0y), (v1x, v1y)) = vs[ct[1][ed], :]
        delx = v1x - v0x
        dely = v1y - v0y
        edge_nodes.append([(v0x+nd*delx, v0y+nd*dely) for nd in nodes1d])

    # subtracts off the value of function at internal nodes times those
    # internal basis functions
    def nodalize(f):
        return f - sum(f.subs(xysub(xx, nd)) * bf
                       for bf, nd in zip(internal_bfs, internal_nodes))

    edge_bfs = []
    if r == 2:
        for ed in range(4):
            lamadj0 = lams[adjacent_edges[ed][0]]
            lamadj1 = lams[adjacent_edges[ed][1]]
            ephi = lamadj0 * lamadj1 * Rs[ed]
            phi = nodalize(ephi) / ephi.subs(xysub(xx, xstars[ed]))
            edge_bfs.append([phi])
    else:
        for ed in range(4):
            ((v0x, v0y), (v1x, v1y)) = vs[ct[1][ed], :]
            Rcur = tunnel_R_edges[ed]
            lam_op = lams[opposite_edges[ed]]

            edge_bfs_cur = []

            for i in range(len(nodes1d)):
                # strike out i:th node
                idcs = [j for j in range(len(nodes1d)) if i != j]
                nodes1d_cur = [nodes1d[j] for j in idcs]
                edge_nodes_cur = [edge_nodes[ed][j]
                                  for j in idcs]

                # construct the 1d interpolation with remaining nodes
                pvals = []
                for nd in edge_nodes_cur:
                    sub = xysub(xx, nd)
                    pval_cur = (-1 * Rcur.subs(sub)**(r-2)
                                / lam_op.subs(sub))
                    pvals.append(pval_cur)

                ptilde = newton_poly(nodes1d_cur, pvals, s)
                xt = xx @ ts[ed]
                vt0 = numpy.asarray((v0x, v0y)) @ ts[ed]
                vt1 = numpy.asarray((v1x, v1y)) @ ts[ed]
                p = ptilde.subs({s: (xt-vt0) / (vt1-vt0)})

                prebf = (lams[adjacent_edges[ed][0]]
                         * lams[adjacent_edges[ed][1]]
                         * (lams[opposite_edges[ed]] * p
                            + Rcur**(r-2) * Rs[ed]))

                prebf = nodalize(prebf)
                bfcur = prebf / prebf.subs(xysub(xx, edge_nodes[ed][i]))
                edge_bfs_cur.append(bfcur)

            edge_bfs.append(edge_bfs_cur)

    # vertex basis functions
    vertex_to_adj_edges = {i: tuple([e for e in ct[1] if i in ct[1][e]])
                           for i in ct[0]}
    vertex_to_off_edges = {i: tuple([e for e in ct[1] if i not in ct[1][e]])
                           for i in ct[0]}

    vertex_bfs = []
    for v in range(4):
        ed0, ed1 = vertex_to_off_edges[v]
        lam0 = lams[ed0]
        lam1 = lams[ed1]

        prebf = lam0 * lam1

        # subtract off edge values
        for adj_ed in vertex_to_adj_edges[v]:
            edge_nodes_cur = edge_nodes[adj_ed]
            edge_bfs_cur = edge_bfs[adj_ed]
            for k, (nd, edbf) in enumerate(zip(edge_nodes_cur, edge_bfs_cur)):
                sb = xysub(xx, nd)
                prebf -= lam0.subs(sb) * lam1.subs(sb) * edbf

        bf = nodalize(prebf) / prebf.subs(xysub(xx, vs[v, :]))
        vertex_bfs.append(bf)

    bfs = vertex_bfs
    for edbfs in edge_bfs:
        bfs.extend(edbfs)
    bfs.extend(internal_bfs)

    nds = [tuple(vs[i, :]) for i in range(4)]
    for ends in edge_nodes:
        nds.extend(ends)
    nds.extend(internal_nodes)

    return vs, xx, numpy.asarray(bfs)


def ds_sym(ct, r, *, vs=None, sp=symbolics):
    """Symbolically Constructs Arbogast's directly defined C^0 serendipity elements,
    which include all polynomials of degree r plus a couple of rational functions.
    :param ct: The cell topology of the reference quadrilateral.
    :param vs: (Optional) coordinates of cell on which to construct the basis.
               If it is None, this function constructs symbols for the vertices.
    :returns: a 3-tuple containing symbols for the physical cell coordinates and the
              physical cell independent variables (e.g. "x" and "y") and a list
              of the four basis functions.
    """
    if r == 1:
        return ds1_sym(ct, vs=vs, sp=sp)
    else:
        return dsr_sym(ct, r, vs=vs, sp=sp)



from functools import partial
from itertools import chain
from operator import add, methodcaller

import FIAT
import gem
import numpy
from gem.utils import cached_property

from finat.finiteelementbase import FiniteElementBase


class EnrichedElement(FiniteElementBase):
    """A finite element whose basis functions are the union of the
    basis functions of several other finite elements."""

    def __new__(cls, elements):
        elements = tuple(chain.from_iterable(e.elements if isinstance(e, EnrichedElement) else (e,) for e in elements))
        if len(elements) == 1:
            return elements[0]
        else:
            self = super().__new__(cls)
            self.elements = elements
            return self

    @cached_property
    def cell(self):
        result, = set(elem.cell for elem in self.elements)
        return result

    @cached_property
    def complex(self):
        return FIAT.reference_element.max_complex(set(elem.complex for elem in self.elements))

    @cached_property
    def degree(self):
        return tree_map(max, *[elem.degree for elem in self.elements])

    @cached_property
    def formdegree(self):
        ks = set(elem.formdegree for elem in self.elements)
        if None in ks:
            return None
        else:
            return max(ks)

    def entity_dofs(self):
        '''Return the map of topological entities to degrees of
        freedom for the finite element.'''
        return concatenate_entity_dofs(self.cell, self.elements,
                                       methodcaller("entity_dofs"))

    @cached_property
    def entity_permutations(self):
        '''Return the map of topological entities to the map of
        orientations to permutation lists for the finite element'''
        return concatenate_entity_permutations(self.elements)

    @cached_property
    def _entity_support_dofs(self):
        return concatenate_entity_dofs(self.cell, self.elements,
                                       methodcaller("entity_support_dofs"))

    def space_dimension(self):
        '''Return the dimension of the finite element space.'''
        return sum(elem.space_dimension() for elem in self.elements)

    @cached_property
    def index_shape(self):
        return (self.space_dimension(),)

    @cached_property
    def value_shape(self):
        '''A tuple indicating the shape of the element.'''
        shape, = set(elem.value_shape for elem in self.elements)
        return shape

    @cached_property
    def fiat_equivalent(self):
        if self.is_mixed:
            # EnrichedElement is actually a MixedElement
            return FIAT.MixedElement([e.element.fiat_equivalent
                                      for e in self.elements], ref_el=self.cell)
        else:
            return FIAT.EnrichedElement(*(e.fiat_equivalent
                                          for e in self.elements))

    @cached_property
    def is_mixed(self):
        # Avoid circular import dependency
        from finat.mixed import MixedSubElement

        return all(isinstance(e, MixedSubElement) for e in self.elements)

    def _compose_evaluations(self, results):
        keys, = set(map(frozenset, results))

        def merge(tables):
            tables = tuple(tables)
            zeta = self.get_value_indices()
            tensors = []
            for elem, table in zip(self.elements, tables):
                beta_i = elem.get_indices()
                tensors.append(gem.ComponentTensor(
                    gem.Indexed(table, beta_i + zeta),
                    beta_i
                ))
            beta = self.get_indices()
            return gem.ComponentTensor(
                gem.Indexed(gem.Concatenate(*tensors), beta),
                beta + zeta
            )
        return {key: merge(result[key] for result in results)
                for key in keys}

    def basis_evaluation(self, order, ps, entity=None, coordinate_mapping=None):
        '''Return code for evaluating the element at known points on the
        reference element.

        :param order: return derivatives up to this order.
        :param ps: the point set object.
        :param entity: the cell entity on which to tabulate.
        '''
        results = [element.basis_evaluation(order, ps, entity, coordinate_mapping=coordinate_mapping)
                   for element in self.elements]
        return self._compose_evaluations(results)

    def point_evaluation(self, order, refcoords, entity=None, coordinate_mapping=None):
        '''Return code for evaluating the element at an arbitrary points on
        the reference element.

        :param order: return derivatives up to this order.
        :param refcoords: GEM expression representing the coordinates
                          on the reference entity.  Its shape must be
                          a vector with the correct dimension, its
                          free indices are arbitrary.
        :param entity: the cell entity on which to tabulate.
        '''
        results = [element.point_evaluation(order, refcoords, entity, coordinate_mapping)
                   for element in self.elements]
        return self._compose_evaluations(results)

    @property
    def mapping(self):
        mappings = set(elem.mapping for elem in self.elements)
        if len(mappings) != 1:
            return None
        else:
            result, = mappings
            return result


def tree_map(f, *args):
    """Like the built-in :py:func:`map`, but applies to a tuple tree."""
    nonleaf, = set(isinstance(arg, tuple) for arg in args)
    if nonleaf:
        ndim, = set(map(len, args))  # asserts equal arity of all args
        return tuple(tree_map(f, *subargs) for subargs in zip(*args))
    else:
        return f(*args)


def concatenate_entity_dofs(ref_el, elements, method):
    """Combine the entity DoFs from a list of elements into a combined
    dict containing the information for the concatenated DoFs of all
    the elements.

    :arg ref_el: the reference cell
    :arg elements: subelement whose DoFs are concatenated
    :arg method: method to obtain the entity DoFs dict
    :returns: concatenated entity DoFs dict
    """
    entity_dofs = {dim: {i: [] for i in entities}
                   for dim, entities in ref_el.get_topology().items()}
    offsets = numpy.cumsum([0] + list(e.space_dimension()
                                      for e in elements), dtype=int)
    for i, d in enumerate(map(method, elements)):
        for dim, dofs in d.items():
            for ent, off in dofs.items():
                entity_dofs[dim][ent] += list(map(partial(add, offsets[i]), off))
    return entity_dofs


def concatenate_entity_permutations(elements):
    """For each dimension, for each entity, and for each possible
    entity orientation, collect the DoF permutation lists from
    entity_permutations dicts of elements and concatenate them.

    :arg elements: subelements whose DoF permutation lists are concatenated
    :returns: entity_permutation dict of the :class:`EnrichedElement` object
        composed of elements.
    """
    permutations = {}
    for element in elements:
        for dim, e_o_p_map in element.entity_permutations.items():
            dim_permutations = permutations.setdefault(dim, {})
            for e, o_p_map in e_o_p_map.items():
                e_dim_permutations = dim_permutations.setdefault(e, {})
                for o, p in o_p_map.items():
                    o_e_dim_permutations = e_dim_permutations.setdefault(o, [])
                    offset = len(o_e_dim_permutations)
                    o_e_dim_permutations += list(offset + q for q in p)
    return permutations



# Copyright (C) 2018 Cyrus Cheng (Imperial College London)
#
# This file is part of FIAT (https://www.fenicsproject.org)
#
# SPDX-License-Identifier:    LGPL-3.0-or-later
#
# Modified by David A. Ham (david.ham@imperial.ac.uk), 2018

from FIAT import finite_element, polynomial_set, dual_set, functional
from FIAT.reference_element import (Point,
                                    DefaultLine,
                                    UFCInterval,
                                    UFCQuadrilateral,
                                    UFCHexahedron,
                                    UFCTriangle,
                                    UFCTetrahedron,
                                    make_affine_mapping,
                                    flatten_reference_cube)
from FIAT.P0 import P0Dual
import numpy as np

hypercube_simplex_map = {Point(): Point(),
                         DefaultLine(): DefaultLine(),
                         UFCInterval(): UFCInterval(),
                         UFCQuadrilateral(): UFCTriangle(),
                         UFCHexahedron(): UFCTetrahedron()}


class DPC0(finite_element.CiarletElement):
    def __init__(self, ref_el):
        flat_el = flatten_reference_cube(ref_el)
        poly_set = polynomial_set.ONPolynomialSet(hypercube_simplex_map[flat_el], 0)
        dual = P0Dual(ref_el)
        # Implement entity_permutations when we handle that for HigherOrderDPC.
        # Currently, orientation_tuples in P0Dual(ref_el).entity_permutations
        # are missing the extrinsic orientation entry (which should always be 0).
        # We should probably just change o_tuple -> (0, ) + o_tuple for non-
        # tensorproduct elements for consistency.
        dual.entity_permutations = None
        degree = 0
        formdegree = ref_el.get_spatial_dimension()  # n-form
        super().__init__(poly_set=poly_set,
                         dual=dual,
                         order=degree,
                         ref_complex=ref_el,
                         formdegree=formdegree)


class DPCDualSet(dual_set.DualSet):
    """The dual basis for DPC elements.  This class works for
    hypercubes of any dimension.  Nodes are point evaluation at
    equispaced points.  This is the discontinuous version where
    all nodes are topologically associated with the cell itself"""

    def __init__(self, ref_el, flat_el, degree):
        entity_ids = {}
        nodes = []

        # Change coordinates here.
        # Vertices of the simplex corresponding to the reference element.
        v_simplex = hypercube_simplex_map[flat_el].get_vertices()
        # Vertices of the reference element.
        v_hypercube = flat_el.get_vertices()
        # For the mapping, first two vertices are unchanged in all dimensions.
        v_ = [v_hypercube[0], v_hypercube[int(-0.5*len(v_hypercube))]]

        # For dimension 1 upwards,
        # take the next vertex and map it to the midpoint of the edge/face it belongs to, and shares
        # with no other points.
        for d in range(1, flat_el.get_dimension()):
            v_.append(tuple(np.asarray(v_hypercube[flat_el.get_dimension() - d] +
                            np.average(np.asarray(v_hypercube[::2]), axis=0))))
        A, b = make_affine_mapping(v_simplex, tuple(v_))  # Make affine mapping to be used later.

        # make nodes by getting points
        # need to do this dimension-by-dimension, facet-by-facet
        top = hypercube_simplex_map[flat_el].get_topology()

        cur = 0
        for dim in sorted(top):
            for entity in sorted(top[dim]):
                pts_cur = hypercube_simplex_map[flat_el].make_points(dim, entity, degree)
                pts_cur = [tuple(np.matmul(A, np.array(x)) + b) for x in pts_cur]
                nodes_cur = [functional.PointEvaluation(flat_el, x)
                             for x in pts_cur]
                nnodes_cur = len(nodes_cur)
                nodes += nodes_cur
                cur += nnodes_cur

        cube_topology = ref_el.get_topology()
        for dim in sorted(cube_topology):
            entity_ids[dim] = {}
            for entity in sorted(cube_topology[dim]):
                entity_ids[dim][entity] = []

        entity_ids[dim][0] = list(range(len(nodes)))
        super().__init__(nodes, ref_el, entity_ids)


class HigherOrderDPC(finite_element.CiarletElement):
    """The DPC finite element.  It is what it is."""

    def __init__(self, ref_el, degree):
        flat_el = flatten_reference_cube(ref_el)
        poly_set = polynomial_set.ONPolynomialSet(hypercube_simplex_map[flat_el], degree)
        dual = DPCDualSet(ref_el, flat_el, degree)
        formdegree = flat_el.get_spatial_dimension()  # n-form
        super().__init__(poly_set=poly_set,
                         dual=dual,
                         order=degree,
                         ref_complex=ref_el,
                         formdegree=formdegree)


def DPC(ref_el, degree):
    if degree == 0:
        return DPC0(ref_el)
    else:
        return HigherOrderDPC(ref_el, degree)



"""Find the maximal complex in a list of cell complexes.
This is a pass-through from FIAT so that FInAT clients
(e.g. tsfc) don't have to directly import FIAT."""

from FIAT.reference_element import max_complex  # noqa: F401



"""This module defines the UFL finite element classes."""
# Copyright (C) 2008-2016 Martin Sandve Alnæs
#
# This file was originally part of UFL (https://www.fenicsproject.org)
#
# SPDX-License-Identifier:    LGPL-3.0-or-later
#
# Modified by Kristian B. Oelgaard
# Modified by Marie E. Rognes 2010, 2012
# Modified by Anders Logg 2014
# Modified by Massimiliano Leoni, 2016
# Modified by Matthew Scroggs, 2023

from ufl.cell import TensorProductCell, as_cell
from finat.ufl.elementlist import canonical_element_description, simplices
from finat.ufl.finiteelementbase import FiniteElementBase
from ufl.utils.formatting import istr


class FiniteElement(FiniteElementBase):
    """The basic finite element class for all simple finite elements."""
    # TODO: Move these to base?
    __slots__ = ("_short_name", "_sobolev_space",
                 "_mapping", "_variant", "_repr")

    def __new__(cls,
                family,
                cell=None,
                degree=None,
                form_degree=None,
                quad_scheme=None,
                variant=None):
        """Intercepts construction to expand CG, DG, RTCE and RTCF spaces on TensorProductCells."""
        if cell is not None:
            cell = as_cell(cell)

        if isinstance(cell, TensorProductCell):
            # Delay import to avoid circular dependency at module load time
            from finat.ufl.enrichedelement import EnrichedElement
            from finat.ufl.hdivcurl import HCurlElement as HCurl
            from finat.ufl.hdivcurl import HDivElement as HDiv
            from finat.ufl.tensorproductelement import TensorProductElement

            family, short_name, degree, reference_value_shape, sobolev_space, mapping, embedded_degree = \
                canonical_element_description(family, cell, degree, form_degree)

            if family in ["RTCF", "RTCE"]:
                cell_h, cell_v = cell.sub_cells()
                if cell_h.cellname() != "interval":
                    raise ValueError(f"{family} is available on TensorProductCell(interval, interval) only.")
                if cell_v.cellname() != "interval":
                    raise ValueError(f"{family} is available on TensorProductCell(interval, interval) only.")

                C_elt = FiniteElement("CG", "interval", degree, variant=variant)
                D_elt = FiniteElement("DG", "interval", degree - 1, variant=variant)

                CxD_elt = TensorProductElement(C_elt, D_elt, cell=cell)
                DxC_elt = TensorProductElement(D_elt, C_elt, cell=cell)

                if family == "RTCF":
                    return EnrichedElement(HDiv(CxD_elt), HDiv(DxC_elt))
                if family == "RTCE":
                    return EnrichedElement(HCurl(CxD_elt), HCurl(DxC_elt))

            elif family == "NCF":
                cell_h, cell_v = cell.sub_cells()
                if cell_h.cellname() != "quadrilateral":
                    raise ValueError(f"{family} is available on TensorProductCell(quadrilateral, interval) only.")
                if cell_v.cellname() != "interval":
                    raise ValueError(f"{family} is available on TensorProductCell(quadrilateral, interval) only.")

                Qc_elt = FiniteElement("RTCF", "quadrilateral", degree, variant=variant)
                Qd_elt = FiniteElement("DQ", "quadrilateral", degree - 1, variant=variant)

                Id_elt = FiniteElement("DG", "interval", degree - 1, variant=variant)
                Ic_elt = FiniteElement("CG", "interval", degree, variant=variant)

                return EnrichedElement(HDiv(TensorProductElement(Qc_elt, Id_elt, cell=cell)),
                                       HDiv(TensorProductElement(Qd_elt, Ic_elt, cell=cell)))

            elif family == "NCE":
                cell_h, cell_v = cell.sub_cells()
                if cell_h.cellname() != "quadrilateral":
                    raise ValueError(f"{family} is available on TensorProductCell(quadrilateral, interval) only.")
                if cell_v.cellname() != "interval":
                    raise ValueError(f"{family} is available on TensorProductCell(quadrilateral, interval) only.")

                Qc_elt = FiniteElement("Q", "quadrilateral", degree, variant=variant)
                Qd_elt = FiniteElement("RTCE", "quadrilateral", degree, variant=variant)

                Id_elt = FiniteElement("DG", "interval", degree - 1, variant=variant)
                Ic_elt = FiniteElement("CG", "interval", degree, variant=variant)

                return EnrichedElement(HCurl(TensorProductElement(Qc_elt, Id_elt, cell=cell)),
                                       HCurl(TensorProductElement(Qd_elt, Ic_elt, cell=cell)))

            elif family == "Q":
                return TensorProductElement(*[FiniteElement("CG", c, degree, variant=variant)
                                              for c in cell.sub_cells()],
                                            cell=cell)

            elif family == "DQ":
                def dq_family(cell):
                    """Doc."""
                    return "DG" if cell.cellname() in simplices else "DQ"
                return TensorProductElement(*[FiniteElement(dq_family(c), c, degree, variant=variant)
                                              for c in cell.sub_cells()],
                                            cell=cell)

            elif family == "DQ L2":
                def dq_family_l2(cell):
                    """Doc."""
                    return "DG L2" if cell.cellname() in simplices else "DQ L2"
                return TensorProductElement(*[FiniteElement(dq_family_l2(c), c, degree, variant=variant)
                                              for c in cell.sub_cells()],
                                            cell=cell)

        return super().__new__(cls)

    def __init__(self,
                 family,
                 cell=None,
                 degree=None,
                 form_degree=None,
                 quad_scheme=None,
                 variant=None):
        """Create finite element.

        :arg family: The finite element family
        :arg cell: The geometric cell
        :arg degree: The polynomial degree (optional)
        :arg form_degree: The form degree (FEEC notation, used when field is
            viewed as k-form)
        :arg quad_scheme: The quadrature scheme (optional)
        :arg variant: Hint for the local basis function variant (optional)

        """
        # Note: Unfortunately, dolfin sometimes passes None for
        # cell. Until this is fixed, allow it:
        if cell is not None:
            cell = as_cell(cell)

        (
            family, short_name, degree, reference_value_shape, sobolev_space, mapping, embedded_degree
        ) = canonical_element_description(family, cell, degree, form_degree)

        # TODO: Move these to base? Might be better to instead
        # simplify base though.
        self._sobolev_space = sobolev_space
        self._mapping = mapping
        self._short_name = short_name or family
        self._variant = variant
        self._embedded_degree = embedded_degree

        # Type check variant
        if variant is not None and not isinstance(variant, str):
            raise ValueError("Illegal variant: must be string or None")

        # Initialize element data
        FiniteElementBase.__init__(self, family, cell, degree, quad_scheme,
                                   reference_value_shape)

        # Cache repr string
        qs = self.quadrature_scheme()
        if qs is None:
            quad_str = ""
        else:
            quad_str = ", quad_scheme=%s" % repr(qs)
        v = self.variant()
        if v is None:
            var_str = ""
        else:
            var_str = ", variant=%s" % repr(v)
        self._repr = "FiniteElement(%s, %s, %s%s%s)" % (
            repr(self.family()), repr(self.cell), repr(self.degree()), quad_str, var_str)
        assert '"' not in self._repr

    def __repr__(self):
        """Format as string for evaluation as Python object."""
        return self._repr

    def _is_globally_constant(self):
        """Doc."""
        return self.family() == "Real"

    def _is_linear(self):
        """Doc."""
        return self.family() == "Lagrange" and self.degree() == 1

    def mapping(self):
        """Return the mapping type for this element ."""
        return self._mapping

    @property
    def sobolev_space(self):
        """Return the underlying Sobolev space."""
        return self._sobolev_space

    def variant(self):
        """Return the variant used to initialise the element."""
        return self._variant

    def reconstruct(self, family=None, cell=None, degree=None, quad_scheme=None, variant=None):
        """Construct a new FiniteElement object with some properties replaced with new values."""
        if family is None:
            family = self.family()
        if cell is None:
            cell = self.cell
        if degree is None:
            degree = self.degree()
        if quad_scheme is None:
            quad_scheme = self.quadrature_scheme()
        if variant is None:
            variant = self.variant()
        return FiniteElement(family, cell, degree, quad_scheme=quad_scheme, variant=variant)

    def __str__(self):
        """Format as string for pretty printing."""
        qs = self.quadrature_scheme()
        qs = "" if qs is None else "(%s)" % qs
        v = self.variant()
        v = "" if v is None else "(%s)" % v
        return "<%s%s%s%s on a %s>" % (self._short_name, istr(self.degree()),
                                       qs, v, self.cell)

    def shortstr(self):
        """Format as string for pretty printing."""
        return f"{self._short_name}{istr(self.degree())}({self.quadrature_scheme()},{istr(self.variant())})"

    def __getnewargs__(self):
        """Return the arguments which pickle needs to recreate the object."""
        return (self.family(),
                self.cell,
                self.degree(),
                None,
                self.quadrature_scheme(),
                self.variant())

    @property
    def embedded_subdegree(self):
        """Return embedded subdegree."""
        subdegree = self.degree()
        if not isinstance(subdegree, int):
            subdegree = min(subdegree)
        if isinstance(self._embedded_degree, int):
            subdegree = min(subdegree, self._embedded_degree)
        return subdegree

    @property
    def embedded_superdegree(self):
        """Return embedded superdegree."""
        superdegree = self.degree()
        if not isinstance(superdegree, int):
            superdegree = max(superdegree)
        if isinstance(self._embedded_degree, int):
            superdegree = max(superdegree, self._embedded_degree)
        return superdegree



from abc import ABCMeta, abstractmethod, abstractproperty
from itertools import chain

import gem
import numpy
from gem.interpreter import evaluate
from gem.optimise import delta_elimination, sum_factorise, traverse_product
from gem.utils import cached_property

from finat.quadrature import make_quadrature


class FiniteElementBase(metaclass=ABCMeta):

    @abstractproperty
    def cell(self):
        '''The reference cell on which the element is defined.'''

    @property
    def complex(self):
        '''The reference cell complex over which bases are defined.
           Can be different than self.cell in the case of macro elements.'''

    @abstractproperty
    def degree(self):
        '''The degree of the embedding polynomial space.

        In the tensor case this is a tuple.
        '''

    @abstractproperty
    def formdegree(self):
        '''Degree of the associated form (FEEC)'''

    @abstractmethod
    def entity_dofs(self):
        '''Return the map of topological entities to degrees of
        freedom for the finite element.'''

    @property
    def entity_permutations(self):
        '''Returns a nested dictionary that gives, for each dimension,
        for each entity, and for each possible entity orientation, the
        DoF permutation array that maps the entity local DoF ordering
        to the canonical global DoF ordering.

        The entity permutations `dict` for the degree 4 Lagrange finite
        element on the interval, for instance, is given by:

        .. code-block:: python3

            {0: {0: {0: [0]},
                 1: {0: [0]}},
             1: {0: {0: [0, 1, 2],
                     1: [2, 1, 0]}}}

        Note that there are two entities on dimension ``0`` (vertices),
        each of which has only one possible orientation, while there is
        a single entity on dimension ``1`` (interval), which has two
        possible orientations representing non-reflected and reflected
        intervals.
        '''
        raise NotImplementedError(f"entity_permutations not yet implemented for {type(self)}")

    @cached_property
    def _entity_closure_dofs(self):
        # Compute the nodes on the closure of each sub_entity.
        entity_dofs = self.entity_dofs()
        return {dim: {e: sorted(chain(*[entity_dofs[d][se]
                                        for d, se in sub_entities]))
                      for e, sub_entities in entities.items()}
                for dim, entities in self.cell.sub_entities.items()}

    def entity_closure_dofs(self):
        '''Return the map of topological entities to degrees of
        freedom on the closure of those entities for the finite
        element.'''
        return self._entity_closure_dofs

    def is_dg(self):
        return self.entity_dofs() == self.entity_closure_dofs()

    @cached_property
    def _entity_support_dofs(self):
        esd = {}
        for entity_dim in self.cell.sub_entities.keys():
            beta = self.get_indices()
            zeta = self.get_value_indices()

            entity_cell = self.cell.construct_subelement(entity_dim)
            quad = make_quadrature(entity_cell, (2*numpy.array(self.degree)).tolist())

            eps = 1.e-8  # Is this a safe value?

            result = {}
            for f in self.entity_dofs()[entity_dim].keys():
                # Tabulate basis functions on the facet
                vals, = self.basis_evaluation(0, quad.point_set, entity=(entity_dim, f)).values()
                # Integrate the square of the basis functions on the facet.
                ints = gem.IndexSum(
                    gem.Product(gem.IndexSum(gem.Product(gem.Indexed(vals, beta + zeta),
                                                         gem.Indexed(vals, beta + zeta)), zeta),
                                quad.weight_expression),
                    quad.point_set.indices
                )
                evaluation, = evaluate([gem.ComponentTensor(ints, beta)])
                ints = evaluation.arr.flatten()
                assert evaluation.fids == ()
                result[f] = [dof for dof, i in enumerate(ints) if i > eps]

            esd[entity_dim] = result
        return esd

    def entity_support_dofs(self):
        '''Return the map of topological entities to degrees of
        freedom that have non-zero support on those entities for the
        finite element.'''
        return self._entity_support_dofs

    @abstractmethod
    def space_dimension(self):
        '''Return the dimension of the finite element space.'''

    @abstractproperty
    def index_shape(self):
        '''A tuple indicating the number of degrees of freedom in the
        element. For example a scalar quadratic Lagrange element on a triangle
        would return (6,) while a vector valued version of the same element
        would return (6, 2)'''

    @abstractproperty
    def value_shape(self):
        '''A tuple indicating the shape of the element.'''

    @property
    def fiat_equivalent(self):
        '''The FIAT element equivalent to this FInAT element.'''
        raise NotImplementedError(
            f"Cannot make equivalent FIAT element for {type(self).__name__}"
        )

    def get_indices(self):
        '''A tuple of GEM :class:`Index` of the correct extents to loop over
        the basis functions of this element.'''

        return tuple(gem.Index(extent=d) for d in self.index_shape)

    def get_value_indices(self):
        '''A tuple of GEM :class:`~gem.Index` of the correct extents to loop over
        the value shape of this element.'''

        return tuple(gem.Index(extent=d) for d in self.value_shape)

    @abstractmethod
    def basis_evaluation(self, order, ps, entity=None, coordinate_mapping=None):
        '''Return code for evaluating the element at known points on the
        reference element.

        :param order: return derivatives up to this order.
        :param ps: the point set object.
        :param entity: the cell entity on which to tabulate.
        :param coordinate_mapping: a
           :class:`~.physically_mapped.PhysicalGeometry` object that
           provides physical geometry callbacks (may be None).
        '''

    @abstractmethod
    def point_evaluation(self, order, refcoords, entity=None, coordinate_mapping=None):
        '''Return code for evaluating the element at an arbitrary points on
        the reference element.

        :param order: return derivatives up to this order.
        :param refcoords: GEM expression representing the coordinates
                          on the reference entity.  Its shape must be
                          a vector with the correct dimension, its
                          free indices are arbitrary.
        :param entity: the cell entity on which to tabulate.
        :param coordinate_mapping: a
           :class:`~.physically_mapped.PhysicalGeometry` object that
           provides physical geometry callbacks (may be None).
        '''

    @property
    def dual_basis(self):
        '''Return a dual evaluation gem weight tensor Q and point set x to dual
        evaluate a function fn at.

        The general dual evaluation is then Q * fn(x) (the contraction of Q
        with fn(x) along the the indices of x and any shape introduced by fn).

        If the dual weights are scalar then Q, for a general scalar FIAT
        element, is a matrix with dimensions

        .. code-block:: text

            (num_nodes, num_points)

        If the dual weights are tensor valued then Q, for a general tensor
        valued FIAT element, is a tensor with dimensions

        .. code-block:: text

            (num_nodes, num_points, dual_weight_shape[0], ..., dual_weight_shape[n])

        If the dual basis is of a tensor product or FlattenedDimensions element
        with N factors then Q in general is a tensor with dimensions

        .. code-block:: text

            (num_nodes_factor1, ..., num_nodes_factorN,
             num_points_factor1, ..., num_points_factorN,
             dual_weight_shape[0], ..., dual_weight_shape[n])

        where num_points_factorX are made free indices that match the free
        indices of x (which is now a TensorPointSet).

        If the dual basis is of a tensor finite element with some shape
        (S1, S2, ..., Sn) then the tensor element tQ is constructed from the
        base element's Q by taking the outer product with appropriately sized
        identity matrices:

        .. code-block:: text

            tQ = Q ⊗ 𝟙ₛ₁ ⊗ 𝟙ₛ₂ ⊗ ... ⊗ 𝟙ₛₙ

        .. note::

            When Q is returned, the contraction indices of the point set are
            already free indices rather than being left in its shape (as either
            ``num_points`` or ``num_points_factorX``). This is to avoid index
            labelling confusion when performing the dual evaluation
            contraction.

        .. note::

            FIAT element dual bases are built from their ``Functional.pt_dict``
            properties. Therefore any FIAT dual bases with derivative nodes
            represented via a ``Functional.deriv_dict`` property does not
            currently have a FInAT dual basis.
        '''
        raise NotImplementedError(
            f"Dual basis not defined for element {type(self).__name__}"
        )

    def dual_evaluation(self, fn):
        '''Get a GEM expression for performing the dual basis evaluation at
        the nodes of the reference element. Currently only works for flat
        elements: tensor elements are implemented in
        :class:`TensorFiniteElement`.

        :param fn: Callable representing the function to dual evaluate.
                   Callable should take in an :class:`AbstractPointSet` and
                   return a GEM expression for evaluation of the function at
                   those points.
        :returns: A tuple ``(dual_evaluation_gem_expression, basis_indices)``
                  where the given ``basis_indices`` are those needed to form a
                  return expression for the code which is compiled from
                  ``dual_evaluation_gem_expression`` (alongside any argument
                  multiindices already encoded within ``fn``)
        '''
        Q, x = self.dual_basis

        expr = fn(x)
        # Apply targeted sum factorisation and delta elimination to
        # the expression
        sum_indices, factors = delta_elimination(*traverse_product(expr))
        expr = sum_factorise(sum_indices, factors)
        # NOTE: any shape indices in the expression are because the
        # expression is tensor valued.
        assert expr.shape == Q.shape[len(Q.shape)-len(expr.shape):]
        shape_indices = gem.indices(len(expr.shape))
        basis_indices = gem.indices(len(Q.shape) - len(expr.shape))
        Qi = Q[basis_indices + shape_indices]
        expri = expr[shape_indices]
        evaluation = gem.IndexSum(Qi * expri, x.indices + shape_indices)
        # Now we want to factorise over the new contraction with x,
        # ignoring any shape indices to avoid hitting the sum-
        # factorisation index limit (this is a bit of a hack).
        # Really need to do a more targeted job here.
        evaluation = gem.optimise.contraction(evaluation, shape_indices)
        return evaluation, basis_indices

    @abstractproperty
    def mapping(self):
        '''Appropriate mapping from the reference cell to a physical cell for
        all basis functions of the finite element.'''


def entity_support_dofs(elem, entity_dim):
    '''Return the map of entity id to the degrees of freedom for which
    the corresponding basis functions take non-zero values.

    :arg elem: FInAT finite element
    :arg entity_dim: Dimension of the cell subentity.
    '''
    return elem.entity_support_dofs()[entity_dim]



"""This module contains an implementation of the COFFEE optimisation
algorithm operating on a GEM representation.

This file is NOT for code generation as a COFFEE AST.
"""

from collections import OrderedDict
import itertools
import logging

import numpy

from gem.gem import IndexSum, one
from gem.optimise import make_sum, make_product
from gem.refactorise import Monomial
from gem.utils import groupby


__all__ = ['optimise_monomial_sum']


def monomial_sum_to_expression(monomial_sum):
    """Convert a monomial sum to a GEM expression.

    :arg monomial_sum: an iterable of :class:`Monomial`s

    :returns: GEM expression
    """
    indexsums = []  # The result is summation of indexsums
    # Group monomials according to their sum indices
    groups = groupby(monomial_sum, key=lambda m: frozenset(m.sum_indices))
    # Create IndexSum's from each monomial group
    for _, monomials in groups:
        sum_indices = monomials[0].sum_indices
        products = [make_product(monomial.atomics + (monomial.rest,)) for monomial in monomials]
        indexsums.append(IndexSum(make_sum(products), sum_indices))
    return make_sum(indexsums)


def index_extent(factor, linear_indices):
    """Compute the product of the extents of linear indices of a GEM expression

    :arg factor: GEM expression
    :arg linear_indices: set of linear indices

    :returns: product of extents of linear indices
    """
    return numpy.prod([i.extent for i in factor.free_indices if i in linear_indices])


def find_optimal_atomics(monomials, linear_indices):
    """Find optimal atomic common subexpressions, which produce least number of
    terms in the resultant IndexSum when factorised.

    :arg monomials: A list of :class:`Monomial`s, all of which should have
                    the same sum indices
    :arg linear_indices: tuple of linear indices

    :returns: list of atomic GEM expressions
    """
    atomics = tuple(OrderedDict.fromkeys(itertools.chain(*(monomial.atomics for monomial in monomials))))

    def cost(solution):
        extent = sum(map(lambda atomic: index_extent(atomic, linear_indices), solution))
        # Prefer shorter solutions, but larger extents
        return (len(solution), -extent)

    optimal_solution = set(atomics)  # pessimal but feasible solution
    solution = set()

    max_it = 1 << 12
    it = iter(range(max_it))

    def solve(idx):
        while idx < len(monomials) and solution.intersection(monomials[idx].atomics):
            idx += 1

        if idx < len(monomials):
            if len(solution) < len(optimal_solution):
                for atomic in monomials[idx].atomics:
                    solution.add(atomic)
                    solve(idx + 1)
                    solution.remove(atomic)
        else:
            if cost(solution) < cost(optimal_solution):
                optimal_solution.clear()
                optimal_solution.update(solution)
            next(it)

    try:
        solve(0)
    except StopIteration:
        logger = logging.getLogger('tsfc')
        logger.warning("Solution to ILP problem may not be optimal: search "
                       "interrupted after examining %d solutions.", max_it)

    return tuple(atomic for atomic in atomics if atomic in optimal_solution)


def factorise_atomics(monomials, optimal_atomics, linear_indices):
    """Group and factorise monomials using a list of atomics as common
    subexpressions. Create new monomials for each group and optimise them recursively.

    :arg monomials: an iterable of :class:`Monomial`s, all of which should have
                    the same sum indices
    :arg optimal_atomics: list of tuples of atomics to be used as common subexpression
    :arg linear_indices: tuple of linear indices

    :returns: an iterable of :class:`Monomials`s after factorisation
    """
    if not optimal_atomics or len(monomials) <= 1:
        return monomials

    # Group monomials with respect to each optimal atomic
    def group_key(monomial):
        for oa in optimal_atomics:
            if oa in monomial.atomics:
                return oa
        assert False, "Expect at least one optimal atomic per monomial."
    factor_group = groupby(monomials, key=group_key)

    # We should not drop monomials
    assert sum(len(ms) for _, ms in factor_group) == len(monomials)

    sum_indices = next(iter(monomials)).sum_indices
    new_monomials = []
    for oa, monomials in factor_group:
        # Create new MonomialSum for the factorised out terms
        sub_monomials = []
        for monomial in monomials:
            atomics = list(monomial.atomics)
            atomics.remove(oa)  # remove common factor
            sub_monomials.append(Monomial((), tuple(atomics), monomial.rest))
        # Continue to factorise the remaining expression
        sub_monomials = optimise_monomials(sub_monomials, linear_indices)
        if len(sub_monomials) == 1:
            # Factorised part is a product, we add back the common atomics then
            # add to new MonomialSum directly rather than forming a product node
            # Retaining the monomial structure enables applying associativity
            # when forming GEM nodes later.
            sub_monomial, = sub_monomials
            new_monomials.append(
                Monomial(sum_indices, (oa,) + sub_monomial.atomics, sub_monomial.rest))
        else:
            # Factorised part is a summation, we need to create a new GEM node
            # and multiply with the common factor
            node = monomial_sum_to_expression(sub_monomials)
            # If the free indices of the new node intersect with linear indices,
            # add to the new monomial as `atomic`, otherwise add as `rest`.
            # Note: we might want to continue to factorise with the new atomics
            # by running optimise_monoials twice.
            if set(linear_indices) & set(node.free_indices):
                new_monomials.append(Monomial(sum_indices, (oa, node), one))
            else:
                new_monomials.append(Monomial(sum_indices, (oa, ), node))
    return new_monomials


def optimise_monomial_sum(monomial_sum, linear_indices):
    """Choose optimal common atomic subexpressions and factorise a
    :class:`MonomialSum` object to create a GEM expression.

    :arg monomial_sum: a :class:`MonomialSum` object
    :arg linear_indices: tuple of linear indices

    :returns: factorised GEM expression
    """
    groups = groupby(monomial_sum, key=lambda m: frozenset(m.sum_indices))
    new_monomials = []
    for _, monomials in groups:
        new_monomials.extend(optimise_monomials(monomials, linear_indices))
    return monomial_sum_to_expression(new_monomials)


def optimise_monomials(monomials, linear_indices):
    """Choose optimal common atomic subexpressions and factorise an iterable
    of monomials.

    :arg monomials: a list of :class:`Monomial`s, all of which should have
                    the same sum indices
    :arg linear_indices: tuple of linear indices

    :returns: an iterable of factorised :class:`Monomials`s
    """
    assert len(set(frozenset(m.sum_indices) for m in monomials)) <= 1, \
        "All monomials required to have same sum indices for factorisation"

    result = [m for m in monomials if not m.atomics]  # skipped monomials
    active_monomials = [m for m in monomials if m.atomics]
    optimal_atomics = find_optimal_atomics(active_monomials, linear_indices)
    result += factorise_atomics(active_monomials, optimal_atomics, linear_indices)
    return result



# Copyright (C) 2013 Andrew T. T. McRae, 2015-2016 Jan Blechta
#
# This file is part of FIAT (https://www.fenicsproject.org)
#
# SPDX-License-Identifier:    LGPL-3.0-or-later

import numpy as np
import math

from FIAT.expansions import polynomial_entity_ids
from FIAT.polynomial_set import PolynomialSet
from FIAT.dual_set import DualSet
from FIAT.finite_element import CiarletElement
from FIAT.barycentric_interpolation import LagrangeLineExpansionSet

__all__ = ['NodalEnrichedElement']


class NodalEnrichedElement(CiarletElement):
    """NodalEnriched element is a direct sum of a sequence of
    finite elements. Primal basis is reorthogonalized to the
    dual basis for nodality.

    The following is equivalent:
        * the constructor is well-defined,
        * the resulting element is unisolvent and its basis is nodal,
        * the supplied elements are unisolvent with nodal basis and
          their primal bases are mutually linearly independent,
        * the supplied elements are unisolvent with nodal basis and
          their dual bases are mutually linearly independent.
    """

    def __init__(self, *elements):

        # Test elements are nodal
        if not all(e.is_nodal() for e in elements):
            raise ValueError("Not all elements given for construction "
                             "of NodalEnrichedElement are nodal")

        # Extract common data
        embedded_degrees = [e.get_nodal_basis().get_embedded_degree() for e in elements]
        embedded_degree = max(embedded_degrees)
        degree = max(e.degree() for e in elements)
        order = max(e.get_order() for e in elements)
        formdegree = None if any(e.get_formdegree() is None for e in elements) \
            else max(e.get_formdegree() for e in elements)
        # LagrangeExpansionSet has fixed degree, ensure we grab the highest one
        elem = elements[embedded_degrees.index(embedded_degree)]
        ref_el = elem.get_reference_complex()
        expansion_set = elem.get_nodal_basis().get_expansion_set()
        mapping = elem.mapping()[0]
        value_shape = elem.value_shape()

        # Sanity check
        assert all(e.get_reference_complex() == ref_el for e in elements)
        assert all(set(e.mapping()) == {mapping, } for e in elements)
        assert all(e.value_shape() == value_shape for e in elements)

        # Merge polynomial sets
        if isinstance(expansion_set, LagrangeLineExpansionSet):
            # Obtain coefficients via interpolation
            points = expansion_set.get_points()
            coeffs = np.vstack([e.tabulate(0, points)[(0,)] for e in elements])
        else:
            assert all(e.get_nodal_basis().get_expansion_set() == expansion_set
                       for e in elements)
            coeffs = [e.get_coeffs() for e in elements]
            coeffs = _merge_coeffs(coeffs, ref_el, embedded_degrees, expansion_set.continuity)
        poly_set = PolynomialSet(ref_el,
                                 degree,
                                 embedded_degree,
                                 expansion_set,
                                 coeffs)

        # Renumber dof numbers
        offsets = np.cumsum([0] + [e.space_dimension() for e in elements[:-1]])
        entity_ids = _merge_entity_ids((e.entity_dofs() for e in elements),
                                       offsets)

        # Merge dual bases
        nodes = [node for e in elements for node in e.dual_basis()]
        ref_el = ref_el.get_parent() or ref_el
        dual_set = DualSet(nodes, ref_el, entity_ids)

        # CiarletElement constructor adjusts poly_set coefficients s.t.
        # dual_set is really dual to poly_set
        super().__init__(poly_set, dual_set, order, formdegree=formdegree, mapping=mapping)


def _merge_coeffs(coeffss, ref_el, degrees, continuity):
    # Indices of the hierachical expansion set on each facet
    entity_ids = polynomial_entity_ids(ref_el, max(degrees), continuity)

    # Number of bases members
    total_dim = sum(c.shape[0] for c in coeffss)

    # Value shape
    value_shape = coeffss[0].shape[1:-1]
    assert all(c.shape[1:-1] == value_shape for c in coeffss)

    # Number of expansion polynomials
    max_expansion_dim = max(c.shape[-1] for c in coeffss)

    # Compose new coeffs
    shape = (total_dim, *value_shape, max_expansion_dim)
    new_coeffs = np.zeros(shape, dtype=coeffss[0].dtype)
    counter = 0
    for c, degree in zip(coeffss, degrees):
        ids = []
        if continuity == "C0":
            dims = sorted(entity_ids)
        else:
            dims = (ref_el.get_spatial_dimension(),)
        for dim in dims:
            if continuity == "C0":
                dimPk = math.comb(degree - 1, dim)
            else:
                dimPk = math.comb(degree + dim, dim)
            for entity in sorted(entity_ids[dim]):
                ids.extend(entity_ids[dim][entity][:dimPk])

        num_members = c.shape[0]
        new_coeffs[counter:counter+num_members, ..., ids] = c
        counter += num_members
    assert counter == total_dim
    return new_coeffs


def _merge_entity_ids(entity_ids, offsets):
    ret = {}
    for i, ids in enumerate(entity_ids):
        for dim in ids:
            if dim not in ret:
                ret[dim] = {}
            for entity in ids[dim]:
                if entity not in ret[dim]:
                    ret[dim][entity] = []
                ret[dim][entity].extend(np.array(ids[dim][entity]) + offsets[i])
    return ret



# This file is part of FIAT (https://www.fenicsproject.org)
#
# SPDX-License-Identifier:    LGPL-3.0-or-later
#
# Written by Pablo D. Brubeck (brubeck@protonmail.com), 2024

# This is not quite Guzman-Neilan, but it has 2*dim*(dim+1) dofs and includes
# dim**2-1 extra constraint functionals.  The first (dim+1)**2 basis functions
# are the reference element bfs, but the extra dim**2-1 are used in the
# transformation theory.

from FIAT import finite_element, polynomial_set, expansions
from FIAT.bernardi_raugel import BernardiRaugelSpace, BernardiRaugelDualSet, BernardiRaugel
from FIAT.alfeld_sorokina import AlfeldSorokina
from FIAT.brezzi_douglas_marini import BrezziDouglasMarini
from FIAT.macro import AlfeldSplit
from FIAT.quadrature_schemes import create_quadrature
from FIAT.restricted import RestrictedElement
from FIAT.nodal_enriched import NodalEnrichedElement

import numpy
import math


def GuzmanNeilanSpace(ref_el, order, kind=1, reduced=False):
    r"""Return a basis for the (extended) Guzman-Neilan H1 space.

    Project the extended Bernardi-Raugel space (Pk + FacetBubble)^d
    into C0 Pk(Alfeld)^d with P_{k-1} divergence, preserving its trace.

    :arg ref_el: a simplex
    :arg order: the maximal polynomial degree
    :kwarg kind: kind = 1 gives Pk^d + GN bubbles,
                 kind = 2 gives C0 Pk(Alfeld)^d + GN bubbles.
    :kwarg reduced: Include tangential bubbles if reduced = False.

    :returns: a PolynomialSet basis for the Guzman-Neilan H1 space.
    """
    sd = ref_el.get_spatial_dimension()
    ref_complex = AlfeldSplit(ref_el)
    C0 = polynomial_set.ONPolynomialSet(ref_complex, sd, shape=(sd,), scale=1, variant="bubble")
    B = take_interior_bubbles(C0)
    if sd > 2:
        B = modified_bubble_subspace(B)

    K = ref_complex if kind == 2 else ref_el
    num_bubbles = sd + 1
    if reduced:
        BR = BernardiRaugel(K, order).get_nodal_basis()
        reduced_dim = BR.get_num_members() - (sd-1) * (sd+1)
        BR = BR.take(list(range(reduced_dim)))
    else:
        num_bubbles *= sd
        BR = BernardiRaugelSpace(K, order)

    GN = constant_div_projection(BR, C0, B, num_bubbles)
    return GN


class GuzmanNeilanH1(finite_element.CiarletElement):
    """The Guzman-Neilan H1-conforming (extended) macroelement."""
    def __init__(self, ref_el, order=1, kind=1):
        sd = ref_el.get_spatial_dimension()
        if order >= sd:
            raise ValueError(f"{type(self).__name__} is only defined for order < dim")
        degree = sd
        poly_set = GuzmanNeilanSpace(ref_el, order, kind=kind)
        ref_complex = poly_set.get_reference_element() if kind == 2 else ref_el
        dual = BernardiRaugelDualSet(ref_complex, order, degree=degree)
        formdegree = sd - 1  # (n-1)-form
        super().__init__(poly_set, dual, degree, formdegree, mapping="contravariant piola")


class GuzmanNeilanFirstKindH1(GuzmanNeilanH1):
    """The Guzman-Neilan H1-conforming (extended) macroelement of the first kind.

    Reference element: a simplex of any dimension.
    Function space: Pk^d + normal facet bubbles with div in P0, with 1 <= k < dim.
    Degrees of freedom: evaluation at Pk lattice points, and normal moments on faces.

    This element belongs to a Stokes complex, and is paired with unsplit DG_{k-1}.
    """
    def __init__(self, ref_el, order=1):
        super().__init__(ref_el, order=order, kind=1)


class GuzmanNeilanSecondKindH1(GuzmanNeilanH1):
    """The Guzman-Neilan H1-conforming (extended) macroelement of the second kind.

    Reference element: a simplex of any dimension.
    Function space: C0 Pk^d(Alfeld) + normal facet bubbles with div in P0, with 1 <= k < dim.
    Degrees of freedom: evaluation at Pk(Alfeld) lattice points, and normal moments on faces.

    This element belongs to a Stokes complex, and is paired with DG_{k-1}(Alfeld).
    """
    def __init__(self, ref_el, order=1):
        super().__init__(ref_el, order=order, kind=2)


def GuzmanNeilanH1div(ref_el, degree=2, reduced=False):
    """The Guzman-Neilan H1(div)-conforming (extended) macroelement.

    Reference element: a simplex of any dimension.
    Function space: C0 P2^d(Alfeld) with C0 P1 divergence + normal facet bubbles with div in P0.
    Degrees of freedom: evaluation at P2(Alfeld) lattice points, divergence at P1 lattice points,
                        and normal moments on faces.

    This element belongs to a Stokes complex, and is paired with CG1(Alfeld).
    """
    order = 0
    AS = AlfeldSorokina(ref_el, 2)
    if reduced or ref_el.get_spatial_dimension() <= 2:
        order = 1
        # Only extract the div bubbles
        div_nodes = [i for i, node in enumerate(AS.dual_basis())
                     if len(node.deriv_dict) > 0]
        AS = RestrictedElement(AS, indices=div_nodes)
    GN = GuzmanNeilanH1(ref_el, order=order)
    return NodalEnrichedElement(AS, GN)


def inner(v, u, qwts):
    """Compute the L2 inner product from tabulation arrays and quadrature weights"""
    return numpy.tensordot(numpy.multiply(v, qwts), u,
                           axes=(range(1, v.ndim), range(1, u.ndim)))


def div(U):
    """Compute the divergence from tabulation dict."""
    return sum(U[k][:, k.index(1), :] for k in U if sum(k) == 1)


def take_interior_bubbles(P, degree=None):
    """Extract the interior bubbles up to the given degree from a complete PolynomialSet."""
    ref_complex = P.get_reference_element()
    ncomp = numpy.prod(P.get_shape())
    dimPk = P.expansion_set.get_num_members(P.degree)
    assert ncomp * dimPk == P.get_num_members()
    continuity = P.expansion_set.continuity
    entity_ids = expansions.polynomial_entity_ids(ref_complex, P.degree,
                                                  continuity=continuity)
    if degree is None or degree >= P.degree:
        slices = {dim: slice(None) for dim in entity_ids}
    else:
        slices = {dim: slice(math.comb(degree-1, dim)) for dim in entity_ids}

    ids = [i + j * dimPk
           for dim in slices
           for f in sorted(ref_complex.get_interior_facets(dim))
           for i in entity_ids[dim][f][slices[dim]]
           for j in range(ncomp)]
    return P.take(ids)


def modified_bubble_subspace(B):
    """Construct the interior bubble space M_k(K^r) from (Guzman and Neilan, 2019)."""
    ref_complex = B.get_reference_element()
    sd = ref_complex.get_spatial_dimension()
    degree = B.degree
    rule = create_quadrature(ref_complex, 2*degree)
    qpts, qwts = rule.get_points(), rule.get_weights()

    # tabulate the linear hat function associated with the barycenter
    hat = B.take([0])
    hat_at_qpts = hat.tabulate(qpts)[(0,)*sd][0, 0]

    # tabulate the bubbles = hat ** (degree - k) * BDMk_facet
    ref_el = ref_complex.get_parent()
    bubbles = [numpy.eye(sd)[:, :, None] * hat_at_qpts[None, None, :] ** degree]
    for k in range(1, degree):
        # tabulate the BDM facet functions
        BDM = BrezziDouglasMarini(ref_el, k)
        BDM_facet = BDM.get_nodal_basis().take(BDM.dual.get_indices("facet"))
        phis = BDM_facet.tabulate(qpts)[(0,)*sd]

        bubbles.append(numpy.multiply(phis, hat_at_qpts ** (degree-k)))

    bubbles = numpy.concatenate(bubbles, axis=0)

    # store the bubbles into a PolynomialSet via L2 projection
    v = B.tabulate(qpts)[(0,) * sd]
    coeffs = numpy.linalg.solve(inner(v, v, qwts), inner(v, bubbles, qwts))
    coeffs = numpy.tensordot(coeffs, B.get_coeffs(), axes=(0, 0))
    M = polynomial_set.PolynomialSet(ref_complex, degree, degree,
                                     B.get_expansion_set(), coeffs)
    return M


def constant_div_projection(BR, C0, M, num_bubbles):
    """Project the BR space into C0 Pk(Alfeld)^d with P_{k-1} divergence."""
    ref_complex = C0.get_reference_element()
    sd = ref_complex.get_spatial_dimension()
    degree = C0.degree
    rule = create_quadrature(ref_complex, 2*degree)
    qpts, qwts = rule.get_points(), rule.get_weights()

    # Take the test space for the divergence in L2 \ R
    Q = polynomial_set.ONPolynomialSet(ref_complex, degree-1)
    Q = Q.take(list(range(1, Q.get_num_members())))
    P = Q.tabulate(qpts)[(0,)*sd]
    P -= numpy.dot(P, qwts)[:, None] / sum(qwts)

    U = M.tabulate(qpts, 1)
    X = BR.tabulate(qpts, 1)
    # Invert the divergence
    B = inner(P, div(U), qwts)
    g = inner(P, div(X)[-num_bubbles:], qwts)
    w = numpy.linalg.solve(B, g)

    # Add correction to BR bubbles
    v = C0.tabulate(qpts)[(0,)*sd]
    coeffs = numpy.linalg.solve(inner(v, v, qwts), inner(v, X[(0,)*sd], qwts))
    coeffs = coeffs.T.reshape(BR.get_num_members(), sd, -1)
    coeffs[-num_bubbles:] -= numpy.tensordot(w, M.get_coeffs(), axes=(0, 0))
    GN = polynomial_set.PolynomialSet(ref_complex, degree, degree,
                                      C0.get_expansion_set(), coeffs)
    return GN
